import datetime
import os
import pytz
import requests

from typing import Any, Optional, List, Dict, Union, Text, Tuple, Callable
from jsonschema import validate

from pykechain.defaults import API_EXTRA_PARAMS
from pykechain.enums import WidgetTypes, Category, WidgetTitleValue
from pykechain.exceptions import APIError, IllegalArgumentError, NotFoundError
from pykechain.models import BaseInScope
from pykechain.models.widgets.enums import MetaWidget, AssociatedObjectId
from pykechain.models.widgets.helpers import _set_title, TITLE_TYPING
from pykechain.models.widgets.widget_schemas import widget_meta_schema
from pykechain.utils import Empty, empty, slugify_ref


class Widget(BaseInScope):
    """A virtual object representing a KE-chain Widget.

    :cvar basestring id: UUID of the widget
    :cvar basestring title: Title of the widget
    :cvar basestring ref: Reference of the widget
    :cvar basestring widget_type: Type of the widget. Should be one of :class:`WidgetTypes`
    :cvar dict meta: Meta definition of the widget
    :cvar int order: Order of the widget in the list of widgets
    :cvar bool has_subwidgets: if the widgets contains any subwidgets. In case this widget being eg. a Multicolumn
    :cvar float progress: Progress of the widget
    """

    schema = widget_meta_schema

    def __init__(self, json: Dict, manager: 'WidgetsManager' = None, **kwargs) -> None:
        """Construct a Widget from a KE-chain 2 json response.

        :param json: the json response to construct the :class:`Part` from
        :type json: dict
        """
        # we need to run the init of 'Base' instead of 'Part' as we do not need the instantiation of properties
        super(Widget, self).__init__(json, **kwargs)
        del self.name

        self.title = json.get("title")
        self.ref = json.get("ref")
        self.manager = manager

        self.widget_type = json.get('widget_type')
        # set schema
        if self._client:
            self.schema = self._client.widget_schema(self.widget_type)

        self.meta = self.validate_meta(json.get('meta'))
        self.order = json.get('order')
        self._activity_id = json.get('activity_id')
        self._parent_id = json.get('parent_id')
        self.has_subwidgets = json.get('has_subwidgets')
        self._scope_id = json.get('scope_id')  # TODO duplicate with `scope_id` attribute
        self.progress = json.get('progress')

    def __repr__(self):  # pragma: no cover
        return "<pyke {} '{}' id {}>".format(self.__class__.__name__, self.widget_type, self.id[-8:])

    @property
    def title_visible(self) -> Optional[Text]:
        """
        Return the title of the widget displayed in KE-chain.

        :return: title string
        :rtype str
        """
        show_title_value = self.meta.get(MetaWidget.SHOW_TITLE_VALUE)
        if show_title_value == WidgetTitleValue.NO_TITLE:
            return None
        elif show_title_value == WidgetTitleValue.CUSTOM_TITLE:
            return self.meta.get(MetaWidget.CUSTOM_TITLE)

        elif show_title_value == WidgetTitleValue.DEFAULT:
            try:
                if self.widget_type == WidgetTypes.PROPERTYGRID:
                    return self._client.part(pk=self.meta.get(AssociatedObjectId.PART_INSTANCE_ID)).name
                elif self.widget_type in [WidgetTypes.FILTEREDGRID, WidgetTypes.SUPERGRID]:
                    return self._client.part(pk=self.meta.get(AssociatedObjectId.PART_MODEL_ID), category=None).name
                elif self.widget_type in [WidgetTypes.SERVICE, WidgetTypes.NOTEBOOK]:
                    return self._client.service(pk=self.meta.get(AssociatedObjectId.SERVICE_ID)).name
                elif self.widget_type in [WidgetTypes.ATTACHMENTVIEWER, WidgetTypes.SIGNATURE]:
                    return self._client.property(pk=self.meta.get(AssociatedObjectId.PROPERTY_INSTANCE_ID),
                                                 category=None).name
                elif self.widget_type == WidgetTypes.CARD:
                    return self.scope.name
                else:
                    # TODO Weather, Scope and Task widgets display type in user's language: retrieve from locize API?
                    #  https://api.locize.app/b3df49f9-caf4-4282-9785-43113bff1ff7/latest/nl/common
                    return None

            except NotFoundError:  # pragma: no cover
                return None
        else:  # pragma: no cover
            return None

    def activity(self) -> 'Activity':
        """Activity associated to the widget.

        :return: The Activity
        :rtype: :class:`Activity`
        """
        return self._client.activity(id=self._activity_id)

    def parent(self) -> 'Widget':
        """Parent widget.

        :return: The parent of this widget.
        :rtype: :class:`Widget`
        """
        if not self._parent_id:
            raise NotFoundError('Widget has no parent widget (parent_id is null).')

        return self._client.widget(pk=self._parent_id)

    def validate_meta(self, meta: Dict) -> Dict:
        """Validate the meta and return the meta if validation is successfull.

        :param meta: meta of the widget to be validated.
        :type meta: dict
        :return meta: if the meta is validated correctly
        :raise: `ValidationError`
        """
        return validate(meta, self.schema) is None and meta

    @classmethod
    def create(cls, json: Dict, **kwargs) -> 'Widget':
        """Create a widget based on the json data.

        This method will attach the right class to a widget, enabling the use of type-specific methods.

        It does not create a widget object in KE-chain. But a pseudo :class:`Widget` object.

        :param json: the json from which the :class:`Widget` object to create
        :type json: dict
        :return: a :class:`Widget` object
        :rtype: :class:`Widget`
        """
        def _type_to_classname(type_widget: str):
            """
            Generate corresponding inner classname based on the widget type.

            :param type_widget: type of the widget (one of :class:`WidgetTypes`)
            :type type_widget: str
            :return: classname corresponding to the widget type
            :rtype: str
            """
            return "{}Widget".format(type_widget.title()) if type_widget else WidgetTypes.UNDEFINED

        widget_type = json.get('widget_type')

        # dispatcher to instantiate the right widget class based on the widget type
        # load all difference widget types from the pykechain.model.widgets module.
        import importlib
        all_widgets = importlib.import_module("pykechain.models.widgets")
        if hasattr(all_widgets, _type_to_classname(widget_type)):
            return getattr(all_widgets, _type_to_classname(widget_type))(json, client=kwargs.pop('client'), **kwargs)
        else:
            return getattr(all_widgets, _type_to_classname(WidgetTypes.UNDEFINED))(json, client=kwargs.pop('client'),
                                                                                   **kwargs)

        #
        # Searchers and retrievers
        #

    def parts(self, *args, **kwargs) -> Any:
        """Retrieve parts belonging to this widget.

        Without any arguments it retrieves the Instances related to this widget only.

        This call only returns the configured properties in an widget. So properties that are not configured
        are not in the returned parts.

        See :class:`pykechain.Client.parts` for additional available parameters.

        """
        return self._client.parts(*args, widget=self.id, **kwargs)

    def associated_parts(self, *args, **kwargs) -> (Any, Any):
        """Retrieve models and instances belonging to this widget.

        This is a convenience method for the :func:`Widget.parts()` method, which is used to retrieve both the
        `Category.MODEL` as well as the `Category.INSTANCE` in a tuple.

        This call only returns the configured (associated) properties in a widget. So properties that are not
        configured (associated) are not in the returned parts.

        If you want to retrieve only the models associated to this task it is better to use:
            `Widget.parts(category=Category.MODEL)`.

        See :func:`pykechain.Client.parts` for additional available parameters.

        :returns: a tuple(models of :class:`PartSet`, instances of :class:`PartSet`)

        """
        models_and_instances = self._client.parts(*args, widget=self.id, category=None, **kwargs)

        models = [p for p in models_and_instances if p.category == Category.MODEL]
        instances = [p for p in models_and_instances if p.category == Category.INSTANCE]

        return models, instances

    #
    # Write methods
    #

    def update_associations(
            self,
            readable_models: Optional[List] = None,
            writable_models: Optional[List] = None,
            part_instance: Optional[Union['Part', Text]] = None,
            parent_part_instance: Optional[Union['Part', Text]] = None,
            **kwargs
    ) -> None:
        """
        Update associations on this widget.

        This is a patch to the list of associations: Existing associations are modified but not removed.

        Alternatively you may use `inputs` or `outputs` as a alias to `readable_models` and `writable_models`
        respectively.

        :param readable_models: list of property models (of :class:`Property` or property_ids (uuids) that has
                                read rights (alias = inputs)
        :type readable_models: List[Property] or List[UUID] or None
        :param writable_models: list of property models (of :class:`Property` or property_ids (uuids) that has
                                write rights (alias = outputs)
        :type writable_models: List[Property] or List[UUID] or None
        :param part_instance: Part object or UUID to be used as instance of the widget
        :type part_instance: Part or UUID
        :param parent_part_instance: Part object or UUID to be used as parent of the widget
        :type parent_part_instance: Part or UUID
        :param kwargs: additional keyword arguments to be passed into the API call as param.
        :return: None
        :raises APIError: when the associations could not be changed
        :raise IllegalArgumentError: when the list is not of the right type
        """
        self._client.update_widget_associations(
            widget=self,
            readable_models=readable_models,
            writable_models=writable_models,
            part_instance=part_instance,
            parent_part_instance=parent_part_instance,
            **kwargs
        )

    def set_associations(
            self,
            readable_models: Optional[List] = None,
            writable_models: Optional[List] = None,
            part_instance: Optional[Union['Part', Text]] = None,
            parent_part_instance: Optional[Union['Part', Text]] = None,
            **kwargs
    ) -> None:
        """
        Set associations on this widget.

        This is an absolute list of associations. If you provide No models, than the associations are cleared.

        Alternatively you may use `inputs` or `outputs` as a alias to `readable_models` and `writable_models`
        respectively.

        :param readable_models: list of property models (of :class:`Property` or property_ids (uuids) that has
                                read rights (alias = inputs)
        :type readable_models: List[Property] or List[UUID] or None
        :param writable_models: list of property models (of :class:`Property` or property_ids (uuids) that has
                                write rights (alias = outputs)
        :type writable_models: List[Property] or List[UUID] or None
        :param part_instance: Part object or UUID to be used as instance of the widget
        :type part_instance: Part or UUID
        :param parent_part_instance: Part object or UUID to be used as parent of the widget
        :type parent_part_instance: Part or UUID
        :param kwargs: additional keyword arguments to be passed into the API call as param.
        :return: None
        :raises APIError: when the associations could not be set
        :raise IllegalArgumentError: when the list is not of the right type
        """
        self._client.set_widget_associations(
            widget=self,
            readable_models=readable_models,
            writable_models=writable_models,
            part_instance=part_instance,
            parent_part_instance=parent_part_instance,
            **kwargs
        )

    def remove_associations(
            self,
            models: List[Union['Property', Text]],
            **kwargs
    ) -> None:
        """
        Remove associated properties from the widget.

        :param models: list of Properties or their uuids
        :return: None
        """
        self._client.remove_widget_associations(
            widget=self,
            models=models,
            **kwargs
        )

    def edit(
            self,
            title: Union[TITLE_TYPING, Empty] = empty,
            meta: Optional[Dict] = None,
            **kwargs
    ) -> None:
        """Edit the details of a widget.

        :param title: New title for the widget
            * False: use the default title, depending on the widget type
            * String value: use this title
            * None: No title at all
        :type title: basestring, None or False
        :param meta: (optional) new Meta definition
        :type meta: dict or None
        :raises APIError: if the widget could not be updated.
        """
        update_dict = dict()

        if meta is not None:
            self.meta.update(meta)
            update_dict.update(dict(meta=self.meta))

        if title is not empty:
            _set_title(meta=self.meta, title=title)
            update_dict.update(meta=self.meta)

        if kwargs:  # pragma: no cover
            update_dict.update(**kwargs)

        url = self._client._build_url('widget', widget_id=self.id)
        response = self._client._request('PUT', url, params=API_EXTRA_PARAMS['widgets'], json=update_dict)

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError("Could not update Widget {}".format(self), response=response)

        self.refresh(json=response.json().get('results')[0])

    def delete(self) -> bool:
        """Delete the widget.

        :return: True when successful
        :rtype: bool
        :raises APIError: when unable to delete the activity
        """
        if self.manager:
            return self.manager.delete_widget(self)
        else:
            self._client.delete_widget(widget=self)

    def copy(self, target_activity: 'Activity', order: Optional[int] = None) -> 'Widget':
        """Copy the widget.

        :param target_activity: `Activity` object under which the desired `Widget` is copied
        :type target_activity: :class:`Activity`
        :param order: (optional) order in the activity of the widget.
        :type order: int or None
        :returns: copied :class:`Widget `
        :raises IllegalArgumentError: if target_activity is not :class:`Activity`

        >>> source_activity = project.activity('Source task')
        >>> target_activity = project.activity('Target task')
        >>> widget_manager = source_activity.widgets()
        >>> widget_to_copy = widget_manager[1]
        >>> widget_to_copy.copy(target_activity=target_activity, order=3)

        """
        from pykechain.models import Activity
        if not isinstance(target_activity, Activity):
            raise IllegalArgumentError("`target_activity` needs to be an activity, got '{}'".format(
                type(target_activity)))

        # Retrieve the widget manager of the target activity
        widget_manager = target_activity.widgets()

        # Get the writable and readable models of the original widget
        associated_part = self.parts(category=Category.MODEL)[0]
        readable_models = list()
        writable_models = list()
        for associated_property in associated_part.properties:
            if associated_property.output:
                writable_models.append(associated_property)
            else:
                readable_models.append(associated_property)

        # Create a perfect copy of the widget
        copied_widget = widget_manager.create_widget(meta=self.meta, widget_type=self.widget_type, title=self.title,
                                                     writable_models=writable_models, readable_models=readable_models,
                                                     order=order)

        return copied_widget

    def move(self, target_activity: 'Activity', order: Optional[int] = None) -> 'Widget':
        """Move the widget.

        :param target_activity: `Activity` object under which the desired `Widget` is moved
        :type target_activity: :class:`Activity`
        :param order: (optional) order in the activity of the widget.
        :type order: int or None
        :returns: copied :class:`Widget `
        :raises IllegalArgumentError: if target_activity is not :class:`Activity`

        """
        moved_widget = self.copy(target_activity=target_activity, order=order)

        self.delete()

        return moved_widget

    @staticmethod
    def _validate_excel_export_inputs(
            target_dir: Text,
            file_name: Text,
            user: 'User',
            default_file_name: Callable,
    ) -> Tuple[
        Text, Text, 'User'
    ]:
        """Check, convert and return the inputs for the Excel exporter functions."""
        if target_dir is None:
            target_dir = os.getcwd()
        elif not isinstance(target_dir, str) or not os.path.isdir(target_dir):
            raise IllegalArgumentError('`target_dir` must be a valid directory.')

        if file_name is None:
            file_name = default_file_name()
        else:
            if not isinstance(file_name, str):
                raise IllegalArgumentError('`file_name` must be a string, "{}" is not.'.format(file_name))
            elif '.xls' in file_name:
                file_name = file_name.split('.xls')[0]
            file_name = slugify_ref(file_name)

        if not file_name.endswith('.xlsx'):
            file_name += '.xlsx'

        from pykechain.models import User
        if user is not None and not isinstance(user, User):
            raise IllegalArgumentError('`user` must be a Pykechain User object, "{}" is not.'.format(user))

        return target_dir, file_name, user

    def download_as_excel(
            self,
            target_dir: Optional[Text] = None,
            file_name: Optional[Text] = None,
            user: 'User' = None,
    ) -> Text:
        """
        Export a grid widget as an Excel sheet.

        :param target_dir: directory (path) to store the Excel sheet.
        :type target_dir: str
        :param file_name: optional, name of the Excel file
        :type file_name: str
        :param user: User object to create timezone-aware datetime values
        :type user: User
        :return: file path of the created Excel sheet
        :rtype str
        """
        grid_widgets = {WidgetTypes.SUPERGRID, WidgetTypes.FILTEREDGRID}
        if self.widget_type not in grid_widgets:
            raise IllegalArgumentError(
                "Only widgets of type {} can be exported to Excel, `{}` is not.".format(grid_widgets, self.widget_type))

        part_model_id = self.meta.get("partModelId")
        parent_instance_id = self.meta.get("parentInstanceId")

        def default_file_name():
            return self._client.model(pk=part_model_id).name if file_name is None else ''

        target_dir, file_name, user = self._validate_excel_export_inputs(target_dir, file_name, user, default_file_name)

        json = dict(
            model_id=part_model_id,
            parent_id=parent_instance_id,
            widget_id=self.id,
            export_format='xlsx',
        )

        if user:
            now_utc = datetime.datetime.now(tz=pytz.utc)

            # Convert UTC time to local timezone based on the user
            timezone_definition = user._json_data['timezone']
            timezone = pytz.timezone(timezone_definition)
            now_local = now_utc.astimezone(timezone)

            offset_minutes = int(round((now_utc - now_local).total_seconds() / 60.0))
        else:
            offset_minutes = 0

        params = dict(
            offset=offset_minutes
        )

        url = self._client._build_url('parts_export')
        response = self._client._request('GET', url, data=json, params=params)

        if response.status_code != requests.codes.ok:  # pragma: no cover
            raise APIError("Could not export widget {}: {}".format(str(response), response.content))

        full_path = os.path.join(target_dir, file_name)

        with open(full_path, 'wb') as f:
            for chunk in response:
                f.write(chunk)

        return full_path



from pykechain.enums import NotificationEvent, NotificationChannels, NotificationStatus
from pykechain.exceptions import NotFoundError, MultipleFoundError, APIError, IllegalArgumentError
from pykechain.models import User, Team
from pykechain.models.notification import Notification
from tests.classes import TestBetamax


class _TestNotification(TestBetamax):
    SUBJECT = '_TEST_SUBJECT'
    MESSAGE = '_TEST_MESSAGE'
    USER_ID = 1

    def setUp(self) -> None:
        super().setUp()
        self.USER = self.client.user(pk=self.USER_ID)
        self.TEAM = self.client.teams()[0]

        for old_notification in self.client.notifications(subject=self.SUBJECT):
            old_notification.delete()


class TestNotificationCreation(_TestNotification):

    def setUp(self):
        super().setUp()
        self.bucket = list()

    def tearDown(self):
        for obj in self.bucket:
            try:
                obj.delete()
            except APIError:
                pass
        super().tearDown()

    def test_create(self):
        # setUp
        notification = self.client.create_notification(subject=self.SUBJECT,
                                                       message=self.MESSAGE)
        self.bucket.append(notification)

        # testing
        self.assertIsInstance(notification, Notification)
        self.assertEqual(self.SUBJECT, notification.subject)
        self.assertEqual(self.MESSAGE, notification.message)

    def test_create_with_inputs(self):
        notification = self.client.create_notification(
            subject=self.SUBJECT,
            message=self.MESSAGE,
            status=NotificationStatus.READY,
            recipients=[self.USER_ID],
            team=self.TEAM,
            from_user=self.USER_ID,
            event=NotificationEvent.EXPORT_ACTIVITY_ASYNC,
            channel=NotificationChannels.EMAIL,
        )
        self.bucket.append(notification)

        self.assertIsInstance(notification, Notification)
        self.assertEqual(NotificationStatus.READY, notification.status)
        self.assertEqual(self.USER_ID, notification.recipient_user_ids[0])
        self.assertEqual(self.TEAM, notification.get_team())
        self.assertEqual(self.USER, notification.get_from_user())
        self.assertEqual(NotificationEvent.EXPORT_ACTIVITY_ASYNC, notification.event)
        self.assertEqual(NotificationChannels.EMAIL, notification.channels[0])

    def test_create_invalid_inputs(self):
        kwargs = dict(subject=self.SUBJECT, message=self.MESSAGE)
        with self.assertRaises(IllegalArgumentError):
            self.bucket.append(self.client.create_notification(subject=False, message=self.MESSAGE))
        with self.assertRaises(IllegalArgumentError):
            self.bucket.append(self.client.create_notification(subject=self.SUBJECT, message=[self.MESSAGE]))
        with self.assertRaises(IllegalArgumentError):
            self.bucket.append(self.client.create_notification(status='sending', **kwargs))
        with self.assertRaises(IllegalArgumentError):
            self.bucket.append(self.client.create_notification(recipients='not a list', **kwargs))
        with self.assertRaises(IllegalArgumentError):
            self.bucket.append(self.client.create_notification(recipients=['not a user id'], **kwargs))
        with self.assertRaises(IllegalArgumentError):
            self.bucket.append(self.client.create_notification(team=0, **kwargs))
        with self.assertRaises(IllegalArgumentError):
            self.bucket.append(self.client.create_notification(from_user='Myself', **kwargs))
        with self.assertRaises(IllegalArgumentError):
            self.bucket.append(self.client.create_notification(event='Update', **kwargs))
        with self.assertRaises(IllegalArgumentError):
            self.bucket.append(self.client.create_notification(channel=[NotificationChannels.EMAIL], **kwargs))

    def test_delete_notification_from_client(self):
        # setUp
        notification = self.client.create_notification(message=self.MESSAGE, subject=self.SUBJECT)
        self.client.delete_notification(notification=notification)

        # testing
        with self.assertRaises(NotFoundError):
            self.client.notification(message=self.MESSAGE, subject=self.SUBJECT)

    def test_delete_notification(self):
        # setUp
        notification = self.client.create_notification(message=self.MESSAGE, subject=self.SUBJECT)
        notification.delete()

        # testing
        with self.assertRaises(NotFoundError):
            self.client.notification(message=self.MESSAGE, subject=self.SUBJECT)


class TestNotifications(_TestNotification):
    def setUp(self):
        super().setUp()
        self.notification = self.client.create_notification(
            subject=self.SUBJECT,
            message=self.MESSAGE,
            recipients=[self.USER_ID],
            team=self.TEAM,
        )

    def tearDown(self):
        self.notification.delete()
        super().tearDown()

    def test_all_notifications_retrieval(self):
        # setUp
        notifications = self.client.notifications()
        number_of_notification = len(notifications)
        dummy_notification = self.client.create_notification(subject="Dummy subject", message="Dummy message")

        notifications_retrieved_again = self.client.notifications()

        # testing
        self.assertTrue(len(notifications_retrieved_again) == number_of_notification + 1)

        # tearDown
        self.client.delete_notification(notification=dummy_notification)

    def test_retrieve_notification(self):
        # testing
        retrieved_notification = self.client.notification(message=self.MESSAGE, subject=self.SUBJECT)

        self.assertIsInstance(retrieved_notification, Notification)
        self.assertEqual(self.notification, retrieved_notification)

    def test_retrieve_notification_raise_not_found(self):
        with self.assertRaises(NotFoundError):
            self.client.notification(message="Doesn't exist")

    def test_retrieve_notification_raise_multiple_found(self):
        # setUp
        clone_testing_notification = self.client.create_notification(subject=self.SUBJECT, message=self.MESSAGE)

        # testing
        with self.assertRaises(MultipleFoundError):
            self.client.notification(message=self.MESSAGE, subject=self.SUBJECT)

        # tearDown
        clone_testing_notification.delete()

    def test_get_recipient_users(self):
        recipients = self.notification.get_recipient_users()

        # testing
        self.assertIsInstance(recipients, list)
        self.assertTrue(recipients)

        first_recipient = recipients[0]

        self.assertIsInstance(first_recipient, User)
        self.assertEqual('superuser', first_recipient.username)

    def test_get_from_user(self):
        from_user = self.notification.get_from_user()

        self.assertTrue(from_user)
        self.assertIsInstance(from_user, User)
        self.assertEqual('pykechain_user', from_user.username)

    def test_get_team(self):
        team = self.notification.get_team()

        self.assertIsInstance(team, Team)
        self.assertEqual(self.TEAM, team)

    def test_edit(self):
        subject = 'NEW SUBJECT'
        message = 'NEW MESSAGE'
        status = NotificationStatus.ARCHIVED
        recipients = [4]
        from_user = 4
        event = NotificationEvent.EXPORT_ACTIVITY_ASYNC
        channel = NotificationChannels.APP

        self.notification.edit(
            subject=subject,
            message=message,
            status=status,
            recipients=recipients,
            team=self.TEAM,
            from_user=from_user,
            event=event,
            channel=channel,
        )

        self.assertEqual(subject, self.notification.subject)
        self.assertEqual(message, self.notification.message)
        self.assertEqual(status, self.notification.status)
        self.assertListEqual(recipients, self.notification.recipient_user_ids)
        self.assertEqual(self.TEAM.id, self.notification.team_id)
        self.assertEqual(from_user, self.notification.from_user_id)
        self.assertEqual(event, self.notification.event)
        self.assertEqual(channel, self.notification.channels[0])

    def test_edit_incorrect_inputs(self):
        with self.assertRaises(IllegalArgumentError):
            self.notification.edit(subject=['Not a string'])
        with self.assertRaises(IllegalArgumentError):
            self.notification.edit(message=False)
        with self.assertRaises(IllegalArgumentError):
            self.notification.edit(status='Deleting')
        with self.assertRaises(IllegalArgumentError):
            self.notification.edit(recipients=True)
        with self.assertRaises(IllegalArgumentError):
            self.notification.edit(recipients=['Not a user ID'])
        with self.assertRaises(IllegalArgumentError):
            self.notification.edit(team=5)
        with self.assertRaises(IllegalArgumentError):
            self.notification.edit(from_user='self')
        with self.assertRaises(IllegalArgumentError):
            self.notification.edit(from_user=5.3)
        with self.assertRaises(IllegalArgumentError):
            self.notification.edit(event='Update')
        with self.assertRaises(IllegalArgumentError):
            self.notification.edit(channel=[NotificationChannels.APP])

    # test added due to #847 - providing no inputs overwrites values
    def test_edit_notification_clear_values(self):
        # setup
        initial_subject = 'NEW SUBJECT'
        initial_message = 'NEW MESSAGE'
        initial_status = NotificationStatus.ARCHIVED
        initial_recipients = [4]
        initial_from_user = 4
        initial_event = NotificationEvent.EXPORT_ACTIVITY_ASYNC
        initial_channel = NotificationChannels.APP

        self.notification.edit(
            subject=initial_subject,
            message=initial_message,
            status=initial_status,
            recipients=initial_recipients,
            team=self.TEAM,
            from_user=initial_from_user,
            event=initial_event,
            channel=initial_channel,
        )

        # Edit without mentioning values, everything should stay the same
        new_subject = "AWESOME SUBJECT NEW"

        self.notification.edit(
            subject=new_subject
        )

        # testing
        self.assertEqual(self.notification.subject, new_subject)
        self.assertEqual(self.notification.message, initial_message)
        self.assertEqual(self.notification.status, initial_status)
        self.assertEqual(self.notification.recipient_user_ids, initial_recipients)
        self.assertEqual(self.notification.from_user_id, initial_from_user)
        self.assertEqual(self.notification.event, initial_event)
        self.assertEqual(self.notification.channels, [initial_channel])

        # Edit with clearing the values, name and status cannot be cleared
        self.notification.edit(
            subject=None,
            message=None,
            status=None,
            recipients=None,
            team=None,
            from_user=None,
            event=None,
            channel=None,
        )

        self.assertEqual(self.notification.subject, new_subject)
        self.assertEqual(self.notification.message, initial_message)
        self.assertEqual(self.notification.status, initial_status)
        self.assertEqual(self.notification.recipient_user_ids, list())
        self.assertEqual(self.notification.from_user_id, None)
        self.assertEqual(self.notification.event, initial_event)
        self.assertEqual(self.notification.channels, [initial_channel])



from pykechain.utils import __dict__inherited__


class Enum(object):
    """Custom enumeration class to support class attributes as options.

    Example
    -------
    >>> class Toppings(Enum):
    ...    CHEESE = "Cheese"
    ...    SALAMI = "Salami"
    >>> topping_choice = Toppings.CHEESE

    """

    @classmethod
    def options(cls):
        """Provide a sorted list of options."""
        return sorted((value, name) for (name, value) in __dict__inherited__(cls=cls, stop=Enum).items())

    @classmethod
    def values(cls):
        """Provide a (sorted) list of values."""
        return [value for (value, name) in cls.options()]


class Multiplicity(Enum):
    """The various multiplicities that are accepted by KE-chain.

    For more information on the representation in KE-chain, please consult the KE-chain `Part documentation`_.

    :cvar ZERO_ONE: Multiplicity 0 to 1
    :cvar ONE: Multiplicity 1
    :cvar ZERO_MANY: Multiplicity 0 to infinity
    :cvar ONE_MANY: Multiplicity 1 to infinity
    """

    ZERO_ONE = "ZERO_ONE"
    ONE = "ONE"
    ZERO_MANY = "ZERO_MANY"
    ONE_MANY = "ONE_MANY"
    # M_N = "M_N"  # not implemented


class Category(Enum):
    """The various categories of Parts that are accepted by KE-chain.

    For more information on the representation in KE-chain, please consult the KE-chain `Part documentation`_.

    :cvar INSTANCE: Category of Instance
    :cvar MODEL: Category of Model
    """

    INSTANCE = "INSTANCE"
    MODEL = "MODEL"


class Classification(Enum):
    """The various classification of Parts that are accepted by KE-chain.

    For more information on the representation in KE-chain, please consult the KE-chain `Part documentation`_.

    :cvar PRODUCT: Classification of the part object is Product
    :cvar CATALOG: Classification of the part object is a CATALOG

    .. _Part documentation: https://support.ke-chain.com/confluence/dosearchsite.action?queryString=concept+part
    """

    PRODUCT = "PRODUCT"
    CATALOG = "CATALOG"


class PropertyType(Enum):
    """The various property types that are accepted by KE-chain.

    For more information on the representation in KE-chain, please consult the KE-chain `Property documentation`_.

    :cvar CHAR_VALUE: a charfield property (single line text)
    :cvar TEXT_VALUE: text property (long text, may span multiple lines)
    :cvar BOOLEAN_VALUE: a boolean value property (True/False)
    :cvar INT_VALUE: integer property (whole number)
    :cvar FLOAT_VALUE: floating point number property (with digits)
    :cvar DATETIME_VALUE: a datetime value property
    :cvar ATTACHMENT_VALUE: an attachment property
    :cvar LINK_VALUE: url property
    :cvar REFERENCE_VALUE: a reference property, a UUID value referring to other part model

    .. versionadded:: 1.14

    :cvar SINGLE_SELECT_VALUE: single select list property (choose from a list)
    :cvar REFERENCES_VALUE: a multi reference property, a list of UUID values referring to other part models

    .. versionadded:: 3.6
    :cvar ACTIVITY_REFERENCES_VALUE: Activity References Property
    :cvar SCOPE_REFERENCES_VALUE: Scope References Property
    :cvar SERVICE_REFERENCES_VALUE: Service Referenes Property
    :cvar TEAM_REFERENCES_VALUE: Team References Property
    :cvar USER_REFERENCES_VALUE: User References Property
    :cvar JSON_VALUE: Generic JSON storage Property
    :cvar GEOJSON_VALUE: GEOJSON property to store map data
    :cvar WEATHER_VALUE: Weather JSON property compatible with the response of weatherapi.com

    .. _Property documentation: https://support.ke-chain.com/confluence/dosearchsite.action?queryString=concept+property
    """

    CHAR_VALUE = "CHAR_VALUE"
    TEXT_VALUE = "TEXT_VALUE"
    BOOLEAN_VALUE = "BOOLEAN_VALUE"
    INT_VALUE = "INT_VALUE"
    FLOAT_VALUE = "FLOAT_VALUE"
    DATETIME_VALUE = "DATETIME_VALUE"
    DATE_VALUE = "DATE_VALUE"
    TIME_VALUE = "TIME_VALUE"
    ATTACHMENT_VALUE = "ATTACHMENT_VALUE"
    LINK_VALUE = "LINK_VALUE"
    SINGLE_SELECT_VALUE = "SINGLE_SELECT_VALUE"
    MULTI_SELECT_VALUE = "MULTI_SELECT_VALUE"
    REFERENCE_VALUE = "REFERENCE_VALUE"
    REFERENCES_VALUE = "REFERENCES_VALUE"
    ACTIVITY_REFERENCES_VALUE = "ACTIVITY_REFERENCES_VALUE"
    SCOPE_REFERENCES_VALUE = "SCOPE_REFERENCES_VALUE"
    SERVICE_REFERENCES_VALUE = "SERVICE_REFERENCES_VALUE"
    TEAM_REFERENCES_VALUE = "TEAM_REFERENCES_VALUE"
    USER_REFERENCES_VALUE = "USER_REFERENCES_VALUE"
    JSON_VALUE = "JSON_VALUE"
    GEOJSON_VALUE = "GEOJSON_VALUE"
    WEATHER_VALUE = "WEATHER_VALUE"


class ActivityType(Enum):
    """The various Activity types that are accepted by KE-chain.

    .. versionadded:: 2.0

    :cvar TASK: a normal task
    :cvar PROCESS: a subprocess (container) containing other tasks
    """

    PROCESS = 'PROCESS'
    TASK = 'TASK'


class ActivityClassification(Enum):
    """The classification of Activities that are accepted by KE-chain.

    .. versionadded:: 2.0
    .. versionchanged:: 3.2
       Add 'APP' environment for KE-chain versions > 3.1


    :cvar WORKFLOW: Classification of the activity is WORKFLOW
    :cvar CATALOG: Classification of the activity is CATALOG
    :cvar APP: Classification of the activity is APP
    """

    WORKFLOW = 'WORKFLOW'
    CATALOG = 'CATALOG'
    APP = 'APP'


class ActivityRootNames(Enum):
    """The classification of Activities that are accepted by KE-chain.

    .. versionadded:: 2.0
    .. versionchanged:: 3.2
       Add 'APP' environment for KE-chain versions > 3.1

    :cvar WORKFLOW_ROOT: Root of the activity is WORKFLOW_ROOT
    :cvar CATALOG_ROOT: Root of the activity is CATALOG_ROOT (below are CATALOG tasks)
    :cvar APP_ROOT: Root of the activity is APP_ROOT (below are APP 'tasks' ie. 'screems')
    """

    WORKFLOW_ROOT = 'WORKFLOW_ROOT'
    CATALOG_ROOT = 'CATALOG_ROOT'
    APP_ROOT = 'APP_ROOT'


activity_root_name_by_classification = {
    ActivityClassification.WORKFLOW: ActivityRootNames.WORKFLOW_ROOT,
    ActivityClassification.CATALOG: ActivityRootNames.CATALOG_ROOT,
    ActivityClassification.APP: ActivityRootNames.APP_ROOT,
}


class WidgetNames(Enum):
    """The various Names of the Widget that can be configured.

    :cvar SUPERGRIDWIDGET: superGridWidget
    :cvar PROPERTYGRIDWIDGET: propertyGridWidget
    :cvar HTMLWIDGET: htmlWidget
    :cvar FILTEREDGRIDWIDGET: filteredGridWidget
    :cvar SERVICEWIDGET: serviceWidget
    :cvar NOTEBOOKWIDGET: notebookWidget
    :cvar ATTACHMENTVIEWERWIDGET: attachmentViewerWidget
    :cvar TASKNAVIGATIONBARWIDGET: taskNavigationBarWidget
    :cvar JSONWIDGET: jsonWidget

    # KE-chain 3 only
    :cvar SIGNATUREWIDGET: signatureWidget
    :cvar CARDWIDGET: cardWidget
    :cvar METAPANELWIDGET: metaPanelWidget
    :cvar MULTICOLUMNWIDGET: multiColumnWidget
    :cvar PROGRESSWIDGET: progressWidget
    :cvar TASKSWIDGET: tasksWidget
    :cvar SERVICECARDWIDGET: serviceCardWidget
    :cvar DASHBOARDWIDGET: 'dashboardWidget'
    """

    SUPERGRIDWIDGET = 'superGridWidget'
    PROPERTYGRIDWIDGET = 'propertyGridWidget'
    HTMLWIDGET = 'htmlWidget'
    FILTEREDGRIDWIDGET = 'filteredGridWidget'
    SERVICEWIDGET = 'serviceWidget'
    NOTEBOOKWIDGET = 'notebookWidget'
    ATTACHMENTVIEWERWIDGET = 'attachmentViewerWidget'
    TASKNAVIGATIONBARWIDGET = 'taskNavigationBarWidget'
    JSONWIDGET = 'jsonWidget'
    METAPANELWIDGET = 'metaPanelWidget'
    MULTICOLUMNWIDGET = 'multiColumnWidget'
    SIGNATUREWIDGET = 'signatureWidget'
    CARDWIDGET = 'cardWidget'
    PROGRESSWIDGET = 'progressWidget'
    TASKSWIDGET = 'taskWidget'
    SERVICECARDWIDGET = 'serviceCardWidget'
    DASHBOARDWIDGET = 'dashboardWidget'


class WidgetTypes(Enum):
    """The various widget types for the widget definitions available to the widget api.

    :cvar UNDEFINED: Undefined Widget
    :cvar PROPERTYGRID: Propertygrid widget
    :cvar SUPERGRID: Supergrid widget
    :cvar HTML: Html widget
    :cvar FILTEREDGRID: Filteredgrid widget
    :cvar SERVICE: Service widget
    :cvar NOTEBOOK: Notebook widget
    :cvar ATTACHMENTVIEWER: Attachmentviewer widget
    :cvar TASKNAVIGATIONBAR: Tasknavigationbar widget
    :cvar JSON: Json widget
    :cvar METAPANEL: Metapanel widget
    :cvar MULTICOLUMN: Multicolumn widget
    :cvar SCOPE: Scope widget
    :cvar THIRDPARTY: Thirdparty widget
    :cvar PROGRESS: Progress widget
    :cvar SIGNATURE: Signature widget
    :cvar CARD: Card widget
    :cvar TASKS: Tasks widget
    :cvar WEATHER: Weather widget
    :cvar SERVICECARD: Servicecard widget
    :cvar DASHBOARD: Dashboard widget
    """

    UNDEFINED = 'UNDEFINED'
    PROPERTYGRID = 'PROPERTYGRID'
    SUPERGRID = 'SUPERGRID'
    HTML = 'HTML'
    FILTEREDGRID = 'FILTEREDGRID'
    SERVICE = 'SERVICE'
    NOTEBOOK = 'NOTEBOOK'
    ATTACHMENTVIEWER = 'ATTACHMENTVIEWER'
    TASKNAVIGATIONBAR = 'TASKNAVIGATIONBAR'
    JSON = 'JSON'
    METAPANEL = 'METAPANEL'
    MULTICOLUMN = 'MULTICOLUMN'
    SCOPE = 'SCOPE'
    THIRDPARTY = 'THIRDPARTY'
    PROGRESS = 'PROGRESS'
    SIGNATURE = 'SIGNATURE'
    CARD = 'CARD'
    TASKS = 'TASKS'
    WEATHER = 'WEATHER'
    SERVICECARD = 'SERVICECARD'
    DASHBOARD = 'DASHBOARD'


WidgetCompatibleTypes = {
    WidgetNames.SUPERGRIDWIDGET: WidgetTypes.SUPERGRID,
    WidgetNames.PROPERTYGRIDWIDGET: WidgetTypes.PROPERTYGRID,
    WidgetNames.HTMLWIDGET: WidgetTypes.HTML,
    WidgetNames.FILTEREDGRIDWIDGET: WidgetTypes.FILTEREDGRID,
    WidgetNames.SERVICEWIDGET: WidgetTypes.SERVICE,
    WidgetNames.NOTEBOOKWIDGET: WidgetTypes.NOTEBOOK,
    WidgetNames.ATTACHMENTVIEWERWIDGET: WidgetTypes.ATTACHMENTVIEWER,
    WidgetNames.TASKNAVIGATIONBARWIDGET: WidgetTypes.TASKNAVIGATIONBAR,
    WidgetNames.JSONWIDGET: WidgetTypes.JSON,
    WidgetNames.METAPANELWIDGET: WidgetTypes.METAPANEL,
    WidgetNames.MULTICOLUMNWIDGET: WidgetTypes.MULTICOLUMN,
    WidgetNames.PROGRESSWIDGET: WidgetTypes.PROGRESS,
    WidgetNames.SIGNATUREWIDGET: WidgetTypes.SIGNATURE,
    WidgetNames.CARDWIDGET: WidgetTypes.CARD,
    WidgetNames.TASKSWIDGET: WidgetTypes.TASKS,
    WidgetNames.SERVICECARDWIDGET: WidgetTypes.SERVICECARD,
    WidgetNames.DASHBOARDWIDGET: WidgetTypes.DASHBOARD
}

default_metapanel_widget = dict(
    name=WidgetNames.METAPANELWIDGET,
    config=dict(),
    meta=dict(
        showAll=True,
    ),
)


class ActivityStatus(Enum):
    """The various Activity statuses that are accepted by KE-chain.

    :cvar OPEN: status of activity is open
    :cvar COMPLETED: status of activity is completed
    """

    OPEN = 'OPEN'
    COMPLETED = 'COMPLETED'


class ScopeStatus(Enum):
    """The various status of a scope.

    .. versionchanged:: 3.0
      The `TEMPLATE` ScopeStatus is deprecated in KE-chain 3

    :cvar ACTIVE: Status of a scope is active (default)
    :cvar CLOSED: Status of a scope is closed
    :cvar TEMPLATE: Status of a scope is a template (not actively used)(deprecated in KE-chain 3.0)
    :cvar DELETING: Status of a scope when the scope is being deleted
    """

    ACTIVE = 'ACTIVE'
    CLOSED = 'CLOSED'
    TEMPLATE = 'TEMPLATE'
    DELETING = 'DELETING'


class ScopeCategory(Enum):
    """The various categories of a scope.

    .. versionadded::3.0

    :cvar LIBRARY_SCOPE: The scope is a library scope
    :cvar USER_SCOPE: The scope is a normal user scope
    :cvar TEMPLATE_SCOPE: The scope is a template scope
    """

    LIBRARY_SCOPE = 'LIBRARY_SCOPE'
    USER_SCOPE = 'USER_SCOPE'
    TEMPLATE_SCOPE = 'TEMPLATE_SCOPE'


class ServiceType(Enum):
    """The file types of sim script.

    :cvar PYTHON_SCRIPT: service is a python script
    :cvar NOTEBOOK: service is a jupyter notebook
    """

    PYTHON_SCRIPT = 'PYTHON SCRIPT'
    NOTEBOOK = 'NOTEBOOK'


class ServiceEnvironmentVersion(Enum):
    """The acceptable versions of python where services run on.

    :cvar PYTHON_3_6: Service execution environment is a python 3.6 container
    :cvar PYTHON_3_7: Service execution environment is a python 3.7 container
    :cvar PYTHON_3_8: Service execution environment is a python 3.8 container
    :cvar PYTHON_3_6_NOTEBOOKS: execution environment is a python 3.6 container with jupyter notebook preinstalled
    :cvar PYTHON_3_8_NOTEBOOKS: execution environment is a python 3.8 container with jupyter notebook preinstalled
    """

    PYTHON_3_6 = '3.6'
    PYTHON_3_7 = '3.7'
    PYTHON_3_8 = '3.8'
    PYTHON_3_6_NOTEBOOKS = '3.6_notebook'
    PYTHON_3_8_NOTEBOOKS = '3.8_notebook'


class ServiceScriptUser(Enum):
    """The acceptable usertypes under which a (trusted) service is run.

    :cvar KENODE_USER: Run as "kenode" user. Equivalent to a manager in a scope.
    :cvar TEAMMANAGER_USER: Run as "kenode_team". Equivalent to a manager in a team. (disabled until available)
    :cvar CONFIGURATOR_USER: Run as "kenode_configurator". Equivalent to GG:Configurator.
    """

    KENODE_USER = "kenode"
    # TEAMMANAGER_USER = "kenode_team"
    CONFIGURATOR_USER = "kenode_configurator"


class ServiceExecutionStatus(Enum):
    """The acceptable states of a running service.

    :cvar LOADING: Execution is in LOADING state (next RUNNING, FAILED)
    :cvar RUNNING: Execution is in RUNNING state (next COMPLETED, FAILED, TERMINATING)
    :cvar COMPLETED: Execution is in COMPLETED state
    :cvar FAILED: Execution is in FAILED state
    :cvar TERMINATING: Execution is in TERMINATING state (next TERMINATED)
    :cvar TERMINATED: Execution is in TERMINATED state
    """

    LOADING = 'LOADING'
    RUNNING = 'RUNNING'
    COMPLETED = 'COMPLETED'
    FAILED = 'FAILED'
    TERMINATING = 'TERMINATING'
    TERMINATED = 'TERMINATED'


class TeamRoles(Enum):
    """Roles that exist for a team member.

    :cvar MEMBER: A normal team member
    :cvar MANAGER: A team member that may manage the team (add or remove members, change team)
    :cvar OWNER: The owner of a team
    """

    MEMBER = "MEMBER"
    MANAGER = "MANAGER"
    OWNER = "OWNER"


class ScopeRoles(Enum):
    """
    Roles that exist for a member of a scope.

    :cvar MANAGER: owner of the scope, has full rights
    :cvar SUPERVISOR: supervisor member of a scope, has the rights as leadmember and rights to manage catalog tasks.
    :cvar LEADMEMBER: elevated member, has assignment rights, no rights on App tasks or Catalog tasks.
    :cvar MEMBER: normal member, only has viewing rights
    """

    MANAGER = "manager"
    SUPERVISOR = "supervisor"
    LEADMEMBER = "leadmember"
    MEMBER = "member"


class ScopeMemberActions(Enum):
    """
    Actions to be performed on the members of a scope.

    :cvar ADD: add a member to the scope
    :cvar REMOVE: delete a member from the scope
    """

    ADD = 'add'
    REMOVE = 'remove'


class KechainEnv(Enum):
    """Environment variables that can be set for pykechain.

    :cvar KECHAIN_URL: full url of KE-chain where to connect to eg: 'https://<some>.ke-chain.com'
    :cvar KECHAIN_TOKEN: authentication token for the KE-chain user provided from KE-chain user account control
    :cvar KECHAIN_USERNAME: the username for the credentials
    :cvar KECHAIN_PASSWORD: the password for the credentials
    :cvar KECHAIN_SCOPE: the name of the project / scope. Should be unique, otherwise use scope_id
    :cvar KECHAIN_SCOPE_ID: the UUID of the project / scope.
    :cvar KECHAIN_FORCE_ENV_USE: set to 'true', '1', 'ok', or 'yes' to always use the environment variables.
    :cvar KECHAIN_SCOPE_STATUS: the status of the Scope to retrieve, defaults to None to retrieve all scopes
    :cvar KECHAIN_CHECK_CERTIFICATES: if the certificates of the URL should be checked.
    """

    KECHAIN_FORCE_ENV_USE = 'KECHAIN_FORCE_ENV_USE'
    KECHAIN_URL = 'KECHAIN_URL'
    KECHAIN_TOKEN = 'KECHAIN_TOKEN'
    KECHAIN_USERNAME = 'KECHAIN_USERNAME'
    KECHAIN_PASSWORD = 'KECHAIN_PASSWORD'
    KECHAIN_SCOPE = 'KECHAIN_SCOPE'
    KECHAIN_SCOPE_ID = 'KECHAIN_SCOPE_ID'
    KECHAIN_SCOPE_STATUS = 'KECHAIN_SCOPE_STATUS'
    KECHAIN_CHECK_CERTIFICATES = 'KECHAIN_CHECK_CERTIFICATES'


class SortTable(Enum):
    """The acceptable sorting options for a grid/table.

    :cvar ASCENDING: Table is sorted in ASCENDING ORDER
    :cvar DESCENDING: Table is sorted in DESCENDING ORDER
    """

    ASCENDING = 'ASC'
    DESCENDING = 'DESC'


class Alignment(Enum):
    """The acceptable alignment options for attachment viewer, navigation bar widgets and service widgets.

    :cvar LEFT: Aligned to the left
    :cvar CENTER: Aligned to the center
    :cvar RIGHT: Aligned to the right
    """

    LEFT = 'left'
    CENTER = 'center'
    RIGHT = 'right'


class NavigationBarAlignment(Alignment):
    """The acceptable alignment options for a Navigation Bar Widget."""

    pass


class PaperSize(Enum):
    """The acceptable paper sizes options for a downloaded PDF.

    :cvar A0: Paper of size A0
    :cvar A1: Paper of size A1
    :cvar A2: Paper of size A2
    :cvar A3: Paper of size A3
    :cvar A4: Paper of size A4
    """

    A0 = 'a0paper'
    A1 = 'a1paper'
    A2 = 'a2paper'
    A3 = 'a3paper'
    A4 = 'a4paper'
    AUTO = 'automatic'


class PaperOrientation(Enum):
    """The acceptable paper orientation options for a downloaded PDF.

    :cvar PORTRAIT: Paper of orientation 'portrait'
    :cvar LANDSCAPE: Paper of orientation 'landscape'
    """

    PORTRAIT = 'portrait'
    LANDSCAPE = 'landscape'


class PropertyVTypes(Enum):
    """The VTypes (or validator types) that are allowed in the json.

    This corresponds to the various validator classes which SHOULD be named:
       `vtype[0].upper() + vtype[1:]`
       eg: 'numbericRangeValidator' has an implementation class of 'NumericRangeValidator'

    .. versionadded:: 2.2

    :cvar NONEVALIDATOR: noneValidator - No validation is done
    :cvar NUMERICRANGE: numericRangeValidator
    :cvar BOOLEANFIELD: booleanFieldValidator
    :cvar REQUIREDFIELD: requiredFieldValidator
    :cvar EVENNUMBER: evenNumberValidator
    :cvar ODDNUMBER: oddNumberValidator
    :cvar REGEXSTRING: regexStringValidator
    :cvar SINGLEREFERENCE: 'singleReferenceValidator'
    :cvar FILEEXTENSION: 'fileExtensionValidator'
    :cvar FILESIZE: 'fileSizeValidator'
    """

    NONEVALIDATOR = 'noneValidator'
    NUMERICRANGE = 'numericRangeValidator'
    BOOLEANFIELD = 'booleanFieldValidator'
    REQUIREDFIELD = 'requiredFieldValidator'
    EVENNUMBER = 'evenNumberValidator'
    ODDNUMBER = 'oddNumberValidator'
    REGEXSTRING = 'regexStringValidator'
    SINGLEREFERENCE = 'singleReferenceValidator'
    FILEEXTENSION = 'fileExtensionValidator'
    FILESIZE = 'fileSizeValidator'

    # fallback
    ALWAYSALLOW = 'alwaysAllowValidator'


class ValidatorEffectTypes(Enum):
    """The effects that can be attached to a validator.

    .. versionadded:: 2.2

    :cvar NONE_EFFECT: noneEffect
    :cvar VISUALEFFECT: visualEffect
    :cvar TEXT_EFFECT: textEffect
    :cvar ERRORTEXT_EFFECT: errorTextEffect
    :cvar HELPTEXT_EFFECT: helpTextEffect
    """

    NONE_EFFECT = 'noneEffect'
    VISUALEFFECT = 'visualEffect'
    TEXT_EFFECT = 'textEffect'
    ERRORTEXT_EFFECT = 'errorTextEffect'
    HELPTEXT_EFFECT = 'helpTextEffect'


class PropertyRepresentation(Enum):
    """
    The Representation configuration to display a property value.

    .. versionadded:: 3.0

    :cvar DECIMAL_PLACES: Amount of decimal places to show the number
    :cvar SIGNIFICANT_DIGITS: Number (count) of significant digits to display the number
    :cvar LINK_TARGET: configuration of a link to open the link in a new browsertab or not.
    :cvar BUTTON: options to represent the choices of a select-list
    """

    DECIMAL_PLACES = "decimalPlaces"
    SIGNIFICANT_DIGITS = "significantDigits"
    LINK_TARGET = "linkTarget"
    BUTTON = 'buttonRepresentation'
    THOUSANDS_SEPARATOR = "thousandsSeparator"
    AUTOFILL = 'autofill'


class OtherRepresentations(Enum):
    """
    Other representations used in KE-chain.

    :cvar CUSTOM_ICON: different font-awesome icons
    """

    CUSTOM_ICON = 'customIcon'


class _AllRepresentations(PropertyRepresentation, OtherRepresentations):
    pass


class ShowColumnTypes(Enum):
    """The columns that can be shown in a Property grid.

    .. versionadded:: 2.3

    :cvar UNIT: unit
    :cvar DESCRIPTION: description
    """

    UNIT = 'unit'
    DESCRIPTION = 'description'


class ScopeWidgetColumnTypes(Enum):
    """The columns that can be shown in a Scope widget grid.

    .. versionadded:: 3.0

    :cvar PROJECT_NAME: Name
    :cvar START_DATE: Start date
    :cvar DUE_DATE: Due date
    :cvar PROGRESS: Progress
    :cvar STATUS: Status
    :cvar TAGS: Tags
    """

    PROJECT_NAME = 'Name'
    START_DATE = 'Start date'
    DUE_DATE = 'Due date'
    PROGRESS = 'Progress'
    STATUS = 'Status'
    TAGS = 'Tags'


class FilterType(Enum):
    """The type of pre-filters that can be set on a Multi Reference Property.

    .. versionadded:: 3.0

    :cvar GREATER_THAN_EQUAL: 'gte'
    :cvar LOWER_THAN_EQUAL: 'lte'
    :cvar CONTAINS: 'icontains'
    :cvar EXACT: 'exact'
    """

    GREATER_THAN_EQUAL = 'gte'
    LOWER_THAN_EQUAL = 'lte'
    CONTAINS = 'icontains'
    EXACT = 'exact'


class ProgressBarColors(Enum):
    """
    Some basic colors that can be set on a Progress Bar inside a Progress Bar Widget.

    .. versionadded:: 3.0

    :cvar BLACK: '#000000'
    :cvar WHITE: '#FFFFFF'
    :cvar RED: 'FF0000'
    :cvar LIME: '#00FF00'
    :cvar BLUE: '#0000FF'
    :cvar YELLOW: '#FFFF00'
    :cvar CYAN: '#00FFFF'
    :cvar MAGENTA: '#FF00FF'
    :cvar SILVER: '#C0C0C0'
    :cvar GRAY: '#808080'
    :cvar MAROON: '#800000'
    :cvar OLIVE: '#808000'
    :cvar GREEN: '#008000'
    :cvar PURPLE: '#800080'
    :cvar TEAL: '#008080'
    :cvar NAVY: '#000080'
    :cvar DEFAULT_COMPLETED: '#339447'
    :cvar DEFAULT_IN_PROGRESS: '#FF6600'
    :cvar DEFAULT_NO_PROGRESS: '#EEEEEE'
    :cvar DEFAULT_IN_PROGRESS_BACKGROUND: '#FC7C3D'
    """

    BLACK = '#000000'
    WHITE = '#FFFFFF'
    RED = '#FF0000'
    LIME = '#00FF00'
    BLUE = '#0000FF'
    YELLOW = '#FFFF00'
    CYAN = '#00FFFF'
    MAGENTA = '#FF00FF'
    SILVER = '#C0C0C0'
    GRAY = '#808080'
    MAROON = '#800000'
    OLIVE = '#808000'
    GREEN = '#008000'
    PURPLE = '#800080'
    TEAL = '#008080'
    NAVY = '#000080'
    DEFAULT_COMPLETED = '#339447'
    DEFAULT_IN_PROGRESS = '#FF6600'
    DEFAULT_NO_PROGRESS = '#EEEEEE'
    DEFAULT_IN_PROGRESS_BACKGROUND = '#FC7C3D'


class LinkTargets(Enum):
    """
    Target for the CardWidget link and Link property representations.

    .. versionadded:: 3.0

    :cvar SAME_TAB: "_self"
    :cvar NEW_TAB: "_blank"
    """

    SAME_TAB = "_self"
    NEW_TAB = '_blank'


class CardWidgetLinkTarget(LinkTargets):
    """Target for the CardWidget, remaining for backwards compatibility."""

    pass


class CardWidgetLinkValue(Enum):
    """
    Link Value for the CardWidget.

    .. versionadded:: 3.0

    :cvar EXTERNAL_LINK: "External link"
    :cvar TASK_LINK: "Task link"
    :cvar NO_LINK: "No link"
    """

    EXTERNAL_LINK = "External link"
    TASK_LINK = "Task link"
    TREE_VIEW = "Tree view"
    NO_LINK = "No link"


class CardWidgetImageValue(Enum):
    """
    Image for the CardWidget.

    .. versionadded:: 3.0

    :cvar CUSTOM_IMAGE: "Custom image"
    :cvar NO_IMAGE: "No image"
    """

    CUSTOM_IMAGE = "Custom image"
    NO_IMAGE = "No image"


class KEChainPages(Enum):
    """
    URL names of built-in KE-chain pages.

    :cvar DETAIL: "detail"
    :cvar TASKS: "activities"
    :cvar WORK_BREAKDOWN: "activitytree"
    :cvar DATA_MODEL: "productmodel"
    :cvar EXPLORER: "product"
    :cvar SERVICES: "scripts"
    """

    DETAIL = 'detail'
    TASKS = 'activities'
    WORK_BREAKDOWN = 'activitytree'
    DATA_MODEL = 'productmodel'
    EXPLORER = 'product'
    SERVICES = 'scripts'
    CATALOG_WBS = 'catalogtree'
    APP_WBS = 'apptree'


KEChainPageLabels = {
    KEChainPages.DETAIL: 'Project details',
    KEChainPages.TASKS: 'Tasks',
    KEChainPages.WORK_BREAKDOWN: 'Work Breakdown',
    KEChainPages.CATALOG_WBS: 'Catalog',
    KEChainPages.APP_WBS: 'App Screens',
    KEChainPages.DATA_MODEL: 'Data model',
    KEChainPages.EXPLORER: 'Explorer',
    KEChainPages.SERVICES: 'Scripts',
}

KEChainPageLabels_nl = {
    KEChainPages.DETAIL: 'Project details',
    KEChainPages.TASKS: 'Taken',
    KEChainPages.WORK_BREAKDOWN: 'Taakverdeling',
    KEChainPages.CATALOG_WBS: 'Catalogus',
    KEChainPages.APP_WBS: 'App schermen',
    KEChainPages.DATA_MODEL: 'Data model',
    KEChainPages.EXPLORER: 'Explorer',
    KEChainPages.SERVICES: 'Scripts',
}

CardWidgetKEChainPageLink = {
    KEChainPages.DETAIL: 'Project',
    KEChainPages.TASKS: 'Tasks',
    KEChainPages.DATA_MODEL: 'Model',
    KEChainPages.EXPLORER: 'Explorer',
    KEChainPages.SERVICES: 'Script',
    KEChainPages.WORK_BREAKDOWN: 'Work Breakdown',
    KEChainPages.CATALOG_WBS: 'Catalog Tasks',
    KEChainPages.APP_WBS: 'App Tasks',
}

KEChainPageIcons = {
    KEChainPages.DETAIL: 'bookmark',
    KEChainPages.TASKS: 'edit',
    KEChainPages.WORK_BREAKDOWN: 'sitemap',
    KEChainPages.CATALOG_WBS: 'books',
    KEChainPages.APP_WBS: 'tablet-alt',
    KEChainPages.DATA_MODEL: 'cube',
    KEChainPages.EXPLORER: 'folder',
    KEChainPages.SERVICES: 'file-code',
}


class SubprocessDisplayMode(Enum):
    """
    URL variations to vary the display of a subprocess activity.

    :cvar ACTIVITIES: "activities"
    :cvar TREEVIEW: "treeview"
    """

    ACTIVITIES = 'activities'
    TREEVIEW = 'treeview'


class URITarget(Enum):
    """
    Side-bar button redirect options.

    :cvar INTERNAL: "internal"
    :cvar EXTERNAL: "external"
    """

    INTERNAL = 'internal'
    EXTERNAL = 'external'


class FontAwesomeMode(Enum):
    """
    Options to display the same icon.

    Source:
    https://fontawesome.com/how-to-use/on-the-web/setup/getting-started

    :cvar SOLID: "solid"
    :cvar REGULAR: "regular"
    :cvar LIGHT: "light"
    """

    SOLID = 'solid'
    REGULAR = 'regular'
    LIGHT = 'light'


class SelectListRepresentations(Enum):
    """
    Options in which a single-select list property options are displayed.

    :cvar DROP_DOWN: "dropdown"
    :cvar CHECK_BOXES: "checkboxes"
    :cvar BUTTONS: "buttons"
    """

    DROP_DOWN = 'dropdown'
    CHECK_BOXES = 'checkboxes'
    BUTTONS = 'buttons'


class ImageFitValue(Enum):
    """
    Options to fit an image on a CardWidget or AttachmentViewerWidget.

    This is a subset from the `object-fit property`_ in HTML.

    :cvar CONTAIN: scale the image to fit within the widget
    :cvar COVER: scale the image to cover the entire widget

    .. _object-fit property: https://developer.mozilla.org/en-US/docs/Web/CSS/object-fit
    """

    CONTAIN = 'contain'
    COVER = 'cover'


class WidgetTitleValue(Enum):
    """
    Options to configure the title of a widget.

    :cvar DEFAULT: Use the default title of the widget type.
    :cvar NO_TITLE: Show no title.
    :cvar CUSTOM_TITLE: Show a custom title text.
    """

    DEFAULT = 'Default'
    NO_TITLE = 'No title'
    CUSTOM_TITLE = 'Custom title'


class NotificationStatus(Enum):
    """
    Options to retrieve a Notification based on its status.

    normal lifecycle:
    - DRAFT, when a message is first saved to the backend and the status is still in draft.
        next states: READY
    - READY: when the message is ready for processing, it is complete and is to be processed
        next states: PROCESSING
    - PROCESSING: when the message is in the process of being send out
        next states: COMPLETED, FAILED
    - COMPLETED: when the message is successfully sent out
        next states: ARCHIVED
    - FAILED: when the message is not successfully sent out
        next states: ARCHIVED
    - ARCHIVED: when the message is archives and waiting for its deletion against a certain retention policy
        next states: None

    :cvar ARCHIVED: "archived" notifications
    :cvar COMPLETED: "completed" notifications
    :cvar DRAFT: "draft" notifications
    :cvar FAILED: "failed" notifications
    :cvar PROCESSING: "processing" notifications
    :cvar READY: "ready" notifications
    """

    ARCHIVED = 'ARCHIVED'
    COMPLETED = 'COMPLETED'
    DRAFT = 'DRAFT'
    FAILED = 'FAILED'
    PROCESSING = 'PROCESSING'
    READY = 'READY'


class NotificationEvent(Enum):
    """
    Options to retrieve a Notification based on its event.

    :cvar SHARE_ACTIVITY_LINK: notifications generated by sharing the link of an `Activity`
    :cvar EXPORT_ACTIVITY_ASYNC: notifications generated by exporting an `Activity`
    :cvar SHARE_ACTIVITY_PDF: notifications generated by sharing the pdf of an `Activity`
    """

    SHARE_ACTIVITY_LINK = 'SHARE_ACTIVITY_LINK'
    EXPORT_ACTIVITY_ASYNC = 'EXPORT_ACTIVITY_ASYNC'
    SHARE_ACTIVITY_PDF = 'SHARE_ACTIVITY_PDF'


class NotificationChannels(Enum):
    """
    Options to retrieve a Notification based on its channel.

    :cvar EMAIL: email notification
    :cvar APP: app notification
    """

    EMAIL = 'EMAIL'
    APP = 'APP'


class LanguageCodes(Enum):
    """
    Options for the language setting of a user.

    :cvar ENGLISH: English
    :cvar FRENCH: French
    :cvar GERMAN: German
    :cvar DUTCH: Dutch
    :cvar ITALIAN: Italian
    """

    ENGLISH = "en"
    FRENCH = "fr"
    GERMAN = "de"
    DUTCH = "nl"
    ITALIAN = "it"



"""All representations.

.. versionadded:: 3.0.x
"""
from .component import RepresentationsComponent
from .representation_base import BaseRepresentation
from .representations import (
    DecimalPlaces,
    SignificantDigits,
    ThousandsSeparator,
    LinkTarget,
    ButtonRepresentation,
    CustomIconRepresentation,
    Autofill,
)

AnyRepresentation = [
    BaseRepresentation,
    # Numbers
    DecimalPlaces,
    SignificantDigits,
    ThousandsSeparator,
    # Link properties
    LinkTarget,
    # Select lists
    ButtonRepresentation,
    # Date, Time, Datetime properties
    Autofill,
    # Scopes and activities
    CustomIconRepresentation,
]

__all__ = [
    AnyRepresentation,
    RepresentationsComponent,
]

rtype_class_map = {r.rtype: r for r in AnyRepresentation}



from pykechain.models import BaseInScope


class Association(BaseInScope):
    """
    A virtual object representing a KE-chain association.

    :ivar property_instance_id
    :ivar property_model_id
    :ivar part_instance_id
    :ivar part_model_id
    :ivar widget_id
    :ivar activity_id
    :ivar writable
    """

    def __init__(self, json, client):
        """Construct an association from provided json data."""
        super().__init__(json=json, client=client)

        self.property_instance_id = json.get('instance_property')
        self.property_model_id = json.get('model_property')
        self.part_instance_id = json.get('instance_part')
        self.part_model_id = json.get('model_part')
        self.widget_id = json.get('widget')
        self.activity_id = json.get('activity')
        self.writable = json.get('writable')
        self.permissions = json.get('permissions')

    def __repr__(self):  # pragma: no cover
        return "<pyke {} id {}>".format(self.__class__.__name__, self.id)



from abc import abstractmethod, ABC
from typing import Union

from jsonschema import validate

from pykechain.enums import PropertyType, Multiplicity, LinkTargets, SelectListRepresentations, FontAwesomeMode
from pykechain.exceptions import IllegalArgumentError, APIError
from pykechain.models import AnyProperty, Property, Scope, Activity, SelectListProperty
from pykechain.models.representations.representation_base import BaseRepresentation
from pykechain.models.representations.representations import ButtonRepresentation, LinkTarget, SignificantDigits, \
    DecimalPlaces, ThousandsSeparator, CustomIconRepresentation, Autofill
from pykechain.models.validators.validator_schemas import options_json_schema
from tests.classes import SixTestCase, TestBetamax

representation_config = dict(
    rtype='buttonRepresentation',
    config=dict(
        buttonRepresentation="dropdown",
    )
)
options = dict(
    representations=[representation_config],
)


class TestRepresentationJSON(SixTestCase):

    def test_valid_button_representation_json(self):
        validate(options_json_schema, options)


class TestRepresentation(SixTestCase):
    def test_create(self):
        representation = BaseRepresentation()

        self.assertIsInstance(representation, BaseRepresentation)

    def test_create_with_value(self):
        representation = DecimalPlaces(value=3)

        self.assertEqual(3, representation.value)

    def test_create_with_object(self):
        empty_prop = Property(json={}, client=None)
        representation = ButtonRepresentation(empty_prop)

        self.assertIsInstance(representation, ButtonRepresentation)

    def test_parse(self):
        empty_slp = SelectListProperty(json={'value_options': {}}, client=None)
        BaseRepresentation.parse(obj=empty_slp, json=representation_config)

    def test_parse_incorrect_rtype(self):
        no_rtype_config = dict(
            config=dict(
                buttonRepresentation="dropdown",
            )
        )
        with self.assertRaises(ValueError):
            BaseRepresentation.parse(obj=None, json=no_rtype_config)

    def test_parse_unknown_rtype(self):
        no_rtype_config = dict(
            rtype='Not a representation type',
            config=dict(
                buttonRepresentation="dropdown",
            )
        )
        with self.assertRaises(TypeError):
            BaseRepresentation.parse(obj=None, json=no_rtype_config)

    def test_component_invalid_object(self):
        empty_activity = Activity(json={'id': '1234567890'}, client=None)
        representation = ThousandsSeparator(empty_activity)

        with self.assertRaises(IllegalArgumentError):
            empty_activity.representations = [representation]

    def test_component_invalid_property_type(self):
        empty_prop = Property(json={'id': '1234567890', 'category': 'MODEL'}, client=None)
        representation = ThousandsSeparator(empty_prop)
        representation.rtype = 'Broken rtype'

        with self.assertRaises(IllegalArgumentError):
            empty_prop.representations = [representation]

    def test_component_not_a_list(self):
        empty_activity = Activity(json={}, client=None)

        with self.assertRaises(IllegalArgumentError):
            empty_activity.representations = 'Howdy!'

    def test_component_not_a_representation(self):
        empty_activity = Activity(json={}, client=None)

        with self.assertRaises(IllegalArgumentError):
            empty_activity.representations = ['Howdy again!']


class Bases:
    """
    Wrapping private test classes in this Base class allows some inheritance mechanisms to be used.

    If the _Test...() classes would not be wrapped, the test framework would try to run these abstract classes.
    """

    class _TestRepresentationLive(TestBetamax):
        """
        Specify these class attributes in every sub-class
        """
        representation_class = None  # type: type(BaseRepresentation)
        value = None
        new_value = None
        incorrect_value = None

        test_object_name = '__test object for representations'

        def setUp(self):
            super().setUp()
            self.obj = self._get_object()
            self.obj.representations = [
                self.representation_class(
                    obj=self.obj,
                    value=self.value,
                ),
            ]

        def tearDown(self):
            if self.obj:
                try:
                    self.obj.delete()
                except APIError:
                    pass
            super().tearDown()

        @abstractmethod
        def _get_object(self) -> Union[Property, Scope, Activity]:
            pass

        def test_create_with_prop(self):
            representation = self.representation_class(prop=self.obj, value=self.value)

            self.assertIsInstance(representation, self.representation_class)

        def test_get_set(self):
            """Test representation accessor property of the Mixin class"""
            reloaded_obj = self.client.reload(self.obj)
            self.assertEqual(self.obj, reloaded_obj)

            # setUp: get representations
            first_representation = reloaded_obj.representations[0]

            self.assertIsInstance(first_representation, self.representation_class)
            self.assertEqual(self.value, first_representation.value)

        def test_set_value(self):
            """Test value accessor property of the Representation class"""
            # Set new value via the representation object
            first_repr = self.obj.representations[0]
            first_repr.value = self.new_value

            # Retrieve representation again
            reloaded_obj = self.client.reload(self.obj)
            self.assertEqual(self.obj, reloaded_obj)

            first_representation = reloaded_obj.representations[0]

            # testing
            self.assertEqual(first_repr.value, first_representation.value)

        def test_unsupported_value(self):

            with self.assertRaises(IllegalArgumentError):
                self.representation_class(
                    obj=self.obj,
                    value=self.incorrect_value,
                )

    class _TestPropertyRepresentation(_TestRepresentationLive):
        property_type = None
        test_model_name = '__test model for representations'
        test_model = None
        incorrect_value = 'Unsupported value'

        def tearDown(self):
            if self.test_model:
                try:
                    self.test_model.delete()
                except APIError:
                    pass
            super().tearDown()

        def _get_object(self):
            parent_model = self.project.model(name='Bike')
            self.test_model = parent_model.add_model(name=self.test_model_name, multiplicity=Multiplicity.ONE)
            obj = self.test_model.add_property(
                name=self.test_object_name,
                property_type=self.property_type,
            )  # type: AnyProperty
            return obj

    class _TestCustomIconRepresentation(_TestRepresentationLive, ABC):
        representation_class = CustomIconRepresentation
        value = 'university'
        new_value = 'bat'
        incorrect_value = ['must be a string']

        def test_set_mode(self):
            representation = self.obj.representations[0]  # type: CustomIconRepresentation

            # Set mode
            representation.display_mode = FontAwesomeMode.SOLID

            reloaded_obj = self.client.reload(obj=self.obj)
            reloaded_repr = reloaded_obj.representations[0]

            # Test get of mode
            self.assertEqual(representation.display_mode, reloaded_repr.display_mode)

        def test_set_mode_incorrect(self):
            representation = self.obj.representations[0]  # type: CustomIconRepresentation

            with self.assertRaises(IllegalArgumentError):
                representation.display_mode = 'fancy colors'


class TestReprSignificantDigits(Bases._TestPropertyRepresentation):
    property_type = PropertyType.FLOAT_VALUE
    representation_class = SignificantDigits
    value = 3
    new_value = 1


class TestReprThousandsSeparator(Bases._TestPropertyRepresentation):
    property_type = PropertyType.FLOAT_VALUE
    representation_class = ThousandsSeparator
    value = None
    new_value = None


class TestReprDecimalPlaces(Bases._TestPropertyRepresentation):
    property_type = PropertyType.FLOAT_VALUE
    representation_class = DecimalPlaces
    value = 3
    new_value = 5


class TestReprLinkTarget(Bases._TestPropertyRepresentation):
    property_type = PropertyType.LINK_VALUE
    representation_class = LinkTarget
    value = LinkTargets.SAME_TAB
    new_value = LinkTargets.NEW_TAB


class TestReprButton(Bases._TestPropertyRepresentation):
    property_type = PropertyType.SINGLE_SELECT_VALUE
    representation_class = ButtonRepresentation
    value = SelectListRepresentations.CHECK_BOXES
    new_value = SelectListRepresentations.BUTTONS

    def setUp(self):
        super().setUp()
        self.obj.options = ['alpha', 'beta', 'gamma', 'omega']


class TestReprAutofill(Bases._TestPropertyRepresentation):
    property_type = PropertyType.DATETIME_VALUE
    representation_class = Autofill
    value = False
    new_value = True


class TestReprActivity(Bases._TestCustomIconRepresentation):

    def _get_object(self):
        return self.project.create_activity(name=self.test_object_name)


class TestReprScope(Bases._TestCustomIconRepresentation):

    def _get_object(self):
        return self.client.create_scope(name=self.test_object_name)
